@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE TemplateHaskell #-}
 -- | Generate @generics-sop@ boilerplate instances using Template Haskell.
 module Generics.SOP.TH
@@ -19,6 +20,7 @@ import Data.Proxy
 import Language.Haskell.TH.Datatype.TyVarBndr
 import Language.Haskell.TH
 import Language.Haskell.TH.Datatype as TH
+import Language.Haskell.TH.Datatype.TyVarBndr
 
 import Generics.SOP.BasicFunctors
 import qualified Generics.SOP.Metadata as SOP
@@ -189,7 +191,7 @@ deriveMetadataType n datatypeInfoName = do
       [ tySynD datatypeInfoName' [] (metadataType' variant name cons) ]
 
 deriveGenericForDataDec ::
-  (Name -> Q Type) -> DatatypeVariant -> Cxt -> Name -> [TyVarBndrUnit] -> [Type] -> [TH.ConstructorInfo] -> Q [Dec]
+  (Name -> Q Type) -> DatatypeVariant -> Cxt -> Name -> [TyVarBndrVis] -> [Type] -> [TH.ConstructorInfo] -> Q [Dec]
 deriveGenericForDataDec f _variant _cxt name _bndrs instTys cons = do
   let typ = appTysSubst f name instTys
   deriveGenericForDataType f typ cons
@@ -204,7 +206,7 @@ deriveGenericForDataType f typ cons = do
   return [inst]
 
 deriveMetadataForDataDec ::
-  (Name -> Q Type) -> DatatypeVariant -> Cxt -> Name -> [TyVarBndrUnit] -> [Type] -> [TH.ConstructorInfo] -> Q [Dec]
+  (Name -> Q Type) -> DatatypeVariant -> Cxt -> Name -> [TyVarBndrVis] -> [Type] -> [TH.ConstructorInfo] -> Q [Dec]
 deriveMetadataForDataDec f variant _cxt name _bndrs instTys cons = do
   let typ = appTysSubst f name instTys
   deriveMetadataForDataType variant name typ cons
@@ -511,7 +513,7 @@ promotedTypeListSubst f (t:ts) = [t| $promotedConsT $(t >>= substType f) $(promo
 appsT :: Name -> [Q Type] -> Q Type
 appsT n = foldl' appT (conT n)
 
-appTyVars :: (Name -> Q Type) -> Name -> [TyVarBndrUnit] -> Q Type
+appTyVars :: (Name -> Q Type) -> Name -> [TyVarBndrVis] -> Q Type
 appTyVars f n bndrs =
   appsT n (map (f . tvName) bndrs)
 
@@ -548,7 +550,7 @@ withDataDec :: TH.DatatypeInfo
                    -- The datatype context
                 -> Name
                    -- The data type's name
-                -> [TyVarBndrUnit]
+                -> [TyVarBndrVis]
                    -- The datatype's type variable binders, both implicit and explicit.
                    -- Examples:
                    --
@@ -580,7 +582,17 @@ withDataDec (TH.DatatypeInfo { datatypeContext   = ctxt
                              , datatypeInstTypes = instTypes
                              , datatypeVariant   = variant
                              , datatypeCons      = cons }) f =
-  f variant ctxt name bndrs instTypes cons
+  checkForTypeData variant $
+  f variant ctxt name (changeTVFlags bndrReq bndrs) instTypes cons
+
+checkForTypeData :: DatatypeVariant -> Q a -> Q a
+checkForTypeData variant q = do
+  case variant of
+#if MIN_VERSION_th_abstraction(0,5,0)
+    TH.TypeData -> fail $ "`type data` declarations not supported"
+#endif
+    _ -> return ()
+  q
 
 checkForGADTs :: TH.ConstructorInfo -> Q a -> Q a
 checkForGADTs (ConstructorInfo { constructorVars    = exVars
@@ -594,3 +606,6 @@ isNewtypeVariant Datatype        = False
 isNewtypeVariant DataInstance    = False
 isNewtypeVariant Newtype         = True
 isNewtypeVariant NewtypeInstance = True
+#if MIN_VERSION_th_abstraction(0,5,0)
+isNewtypeVariant TH.TypeData     = False
+#endif
